%% Benchmark data example 3 - Microbiome Data - Section 5.3
% Estimate zero-sum regression model using constrained lasso  

%% Initial Setup 

% clear workspace
clear; 

%# Load & organize data #%
% load data
zeroSumR = load('data3data.mat');
% extract data
y = zeroSumR.y;
X = zeroSumR.X;
% extract dimensions
[n, p] = size(X);



%% Estimate using constrained lasso solution path algorithm

%# Model setup #%
% set upequality constraints
Aeq = ones(1, p);
beq = 0;
m1 = size(Aeq, 1);  % # of equality constraints
if isempty(Aeq)
    Aeq = zeros(0, p);
    beq = zeros(0, 1);
end


%# Constrained lasso solution path #%
% estimate solution path
[rhosPathClasso, betaHatPath, dfPath] = lsq_classopath(X, y, [], [], ...
    Aeq, beq, 'qp_solver', 'matlab', 'init_method', 'lp');
% scale the tuning parameter to match the zeroSum formulation (which
%   divides the loss fuction by 2n instead of just 2
rhosPath = rhosPathClasso/n;


%# plot solution path #%
% calculate $\ell_1$ norm along path
ell1NormPath = sum(abs(betaHatPath), 1);
% plot solution path
figure;
plot(ell1NormPath(1:end-1), betaHatPath(:, 1:end-1));
title({'Data Example 3: Microbiome Data', ...
    'Solution Path - Constrained Lasso'})
xlabel('$ \|| \widehat{\beta} \||_1$','Interpreter','latex')
ylabel('$\widehat{\beta}$','Interpreter','latex')
xlim([min(ell1NormPath), max(ell1NormPath)]);
set(gca, 'Xdir', 'reverse')
 
%% Choose rho with extended BIC and calculate predicted R-squared

%# Use EBIC to find optimal rho #%
yHatPath = X*betaHatPath;
RSSpath = sum((y - yHatPath).^2);
ebicPath = n*log(RSSpath/n) + 2*log(n)*dfPath;
[~, idxMinEBIC] = min(ebicPath);
rhoStar = rhosPathClasso(idxMinEBIC);
% fitted values at optimal rho
fittedClasso = X*betaHatPath(:, idxMinEBIC);

%# Predicted R-squared #%
% vector to store predicted values
yHatLOOCV = NaN(n, 1);

% LOOCV
for i=1:n
    % estimate model without ith observation
    betaHatLOOCV = lsq_constrsparsereg(X(1:end ~= i, :), y(1:end ~= i), ...
        rhoStar, 'method', 'qp', 'qp_solver', 'gurobi', 'Aeq', Aeq, 'beq', beq);
    % predicted value
    yHatLOOCV(i) = X(i, :)*betaHatLOOCV;
end

% calculate predicted R-squared
PRESS = sum((y - yHatLOOCV).^2);
SST = y'*y;
predR2 = 1- (PRESS/SST);
 
%# save output to file to estimate/graph in R #%
save('data3classo.mat', 'y', 'X', 'rhosPath', 'betaHatPath', ...
    'ell1NormPath', 'fittedClasso', 'predR2')